﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Maze.Helpers;
using Maze.Models.Visual;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Maze.Models
{
    public class Renderable : GameObject
    {
        protected ModelWrapper Model;
        
        //protected Texture2D Diffuse;
        //protected Texture2D Normal;
        //protected Texture2D Specular;
        //public Model Model { get; set; }
        public BoundingBox Hitbox { get; set; }
        public Vector3 Position { get; set; }
        public Matrix Rotation { get; set; }
        public float Scale { get; set; }

        public Renderable(BGame game, string modelname, Vector3? position = null, float rotation = 0, float scale = 0.790f)
            : base(game)
        {
            Model = _game.EnvironmentService.ModelLoader.GetModel(modelname);
            
            if (position == null)
                position = Vector3.Zero;

            Scale = scale;
            
            Position = position.Value;
            Rotation = Matrix.CreateRotationY(MathHelper.ToRadians(rotation));
            //Hitbox = UpdateBoundingBox(Model.Model, _game.CameraService.WorldMatrix);
        }

        public Renderable(BGame game, string modelname, Vector3? position, Matrix rotation, float scale = 0.790f)
            : base(game)
        {
            Model = _game.EnvironmentService.ModelLoader.GetModel(modelname);

            if (position == null)
                position = Vector3.Zero;

            Scale = scale;
            
            Position = position.Value;
            Rotation = rotation;
            //Hitbox = UpdateBoundingBox(Model.Model, _game.CameraService.WorldMatrix);
        }

        public Renderable(BGame game)
            : base(game)
        {
            Position = Vector3.Zero;
            Rotation = Matrix.Identity;
        }

        

        public virtual void Update()
        {

        }

        public virtual void Draw()
        {
            Matrix worldMatrix = Matrix.CreateScale(Scale) *
                Rotation *
                Matrix.CreateTranslation(Position);

            Matrix[] transforms = new Matrix[Model.Model.Bones.Count];
            Model.Model.CopyAbsoluteBoneTransformsTo(transforms);

            var translation = Matrix.CreateTranslation(new Vector3(0, 0, 0));

            Vector3 lightpos = _game.PlayerService.Player.Position;
            lightpos.Y = 100;

            foreach (ModelMesh mesh in Model.Model.Meshes)
            {
                foreach (var effect in mesh.Effects)
                {
                    var w = transforms[mesh.ParentBone.Index] * worldMatrix * translation;

                    effect.Parameters["Diffuse"].SetValue(Model.Diffuse);
                    effect.Parameters["Specular"].SetValue(Model.Specular);
                    effect.Parameters["Normal"].SetValue(Model.Normal);
                    effect.Parameters["Ambient"].SetValue(new Vector3(0.050f, 0.045f, 0.025f));

                    effect.Parameters["light1Dir"].SetValue(Position);
                    effect.Parameters["light1Pos"].SetValue(lightpos);
                    effect.Parameters["light1Color"].SetValue(Color.AliceBlue.ToVector4());
                    effect.Parameters["light1Attenuation"].SetValue(Color.Beige.ToVector4());
                    effect.Parameters["light1Hotspot"].SetValue(43f);
                    effect.Parameters["light1Falloff"].SetValue(45f);

                    effect.Parameters["wvp"].SetValue(w * _game.CameraService.ViewMatrix * _game.CameraService.ProjectionMatrix);
                    effect.Parameters["worldI"].SetValue(Matrix.Invert(w));
                    effect.Parameters["worldIT"].SetValue(Matrix.Transpose(Matrix.Invert(w)));

                    effect.Parameters["viewInv"].SetValue(Matrix.Invert(_game.CameraService.ViewMatrix));

                    effect.Parameters["world"].SetValue(w);

                    foreach (var pass in effect.CurrentTechnique.Passes)
                    {
                        pass.Apply();
                        mesh.Draw();
                    }
                }
            }

            
        }

        public void ChangeModel(string modelname)
        {
            
            Model = _game.EnvironmentService.ModelLoader.GetModel(modelname);
            Hitbox = UpdateBoundingBox(Model.Model, _game.CameraService.WorldMatrix);
            //Model = LoadModel("Models/" + modelname + "/" + modelname);
            //Diffuse = _game.Content.Load<Texture2D>("Models/" + modelname + "/diffuse");
            //Specular = _game.Content.Load<Texture2D>("Models/" + modelname + "/spec");
            //Normal = _game.Content.Load<Texture2D>("Models/" + modelname + "/normal");
        }

        private Model LoadModel(string assetName)
        {
            Model newModel = _game.Content.Load<Model>(assetName); foreach (ModelMesh mesh in newModel.Meshes)
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                    meshPart.Effect = _game.Effect.Clone();
            return newModel;
        }

        protected BoundingBox UpdateBoundingBox(Model model, Matrix worldTransform)
        {
            Matrix worldMatrix = Matrix.CreateScale(Scale) *
                Rotation *
                Matrix.CreateTranslation(Position);


            // Initialize minimum and maximum corners of the bounding box to max and min values
            Vector3 min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
            Vector3 max = new Vector3(float.MinValue, float.MinValue, float.MinValue);

            // For each mesh of the model
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    // Vertex buffer parameters
                    int vertexStride = meshPart.VertexBuffer.VertexDeclaration.VertexStride;
                    int vertexBufferSize = meshPart.NumVertices * vertexStride;

                    // Get vertex data as float
                    float[] vertexData = new float[vertexBufferSize / sizeof(float)];
                    meshPart.VertexBuffer.GetData<float>(vertexData);

                    // Iterate through vertices (possibly) growing bounding box, all calculations are done in world space
                    for (int i = 0; i < vertexBufferSize / sizeof(float); i += vertexStride / sizeof(float))
                    {
                        Vector3 transformedPosition = Vector3.Transform(new Vector3(vertexData[i], vertexData[i + 2], vertexData[i + 1]), Matrix.CreateScale(2f) * worldTransform);

                        min = Vector3.Min(min, transformedPosition);
                        max = Vector3.Max(max, transformedPosition);
                    }
                }
            }

            //min = new Vector3(min.X, min.Z, min.Y);
            //max = new Vector3(max.X, max.Z, max.Y);

            min = Vector3.Transform(min, Rotation);
            max = Vector3.Transform(max, Rotation);

            min = Vector3.Transform(min, Matrix.CreateRotationY(MathHelper.ToRadians(180)));
            max = Vector3.Transform(max, Matrix.CreateRotationY(MathHelper.ToRadians(180)));

            // Create and return bounding box
            return new BoundingBox(Position + min, Position + max);
        }
    }
}
