﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Maze.Models.TileNamespace;
using Maze.Services;
using Microsoft.Xna.Framework;

namespace Maze.Models.SectorNamespace
{
    public class Sector : GameObject, ISector
    {
        
        public Vector3 Key { get { return SectorCoordinate; } }
        public Guid RoomID { get; set; }
        public ISector Preceder { get; set; }
        public List<ISector> Successors { get; set; }
        public IEnumerable<ISector> Neighbours
        {
            get
            {
                if(Preceder != null)
                    yield return Preceder;

                foreach (var successor in Successors)
                {
                    yield return successor;
                }
            }
        }
        public Directions Directions { get; set; }
        public bool IsDeadEnd { get; set; }
        public int DistanceToEnd { get; set; }
        public Vector3 SectorCoordinate { get; set; }
        public Vector3 Coordinate
        {
            get
            {
                return new Vector3(
                           SectorCoordinate.X * 3 * EnvironmentService.TileSize + 96,
                           0,
                           SectorCoordinate.Z * 3 * EnvironmentService.TileSize + 96);
            }
        }

        private List<Tile> Tiles { get; set; }
        private Random _rnd = new Random();
        private bool _isTilesReady = false;

        public Sector(BGame game, Vector3 point) : base(game)
        {
            Tiles = new List<Tile>();
            Successors = new List<ISector>();
            DistanceToEnd = int.MaxValue;

            SectorCoordinate = point;
        }

        public void BuildModels()
        {
            Tiles.Clear();

            string floorType = (this.IsDeadEnd) ? "roomfloor" : "floor";

            for (int i = -1; i <= 1; i++)
            {
                for (int j = -1; j <= 1; j++)
                {
                    Tiles.Add(new Tile(_game, floorType,
                                        new Vector3((Coordinate.X + (i * EnvironmentService.TileSize)), -1,
                                                    (Coordinate.Z + (j * EnvironmentService.TileSize))),
                                        _rnd.Next(4) * 90,
                                        this));
                }
            }

            #region North
            if (!Directions.HasFlag(Directions.North))
            {
                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   180, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   180, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   180, this));
            } 
            #endregion

            #region South
            if (!Directions.HasFlag(Directions.South))
            {
                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   0, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   0, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   0, this));
            }
            #endregion

            #region East
            if (!Directions.HasFlag(Directions.East))
            {
                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   90, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z),
                                   90, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   90, this));
            }
            #endregion

            #region West
            if (!Directions.HasFlag(Directions.West))
            {
                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   -90, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z),
                                   -90, this));

                Tiles.Add(new Tile(_game, "wall",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   -90, this));
            }
            #endregion

            if (Directions.HasFlag(Directions.North) && Directions.HasFlag(Directions.East))
            {
                Tiles.Add(new Tile(_game, "incorner",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   90, this));
            }

            if (Directions.HasFlag(Directions.North) && Directions.HasFlag(Directions.West))
            {
                Tiles.Add(new Tile(_game, "incorner",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z + EnvironmentService.TileSize),
                                   180, this));
            }

            if (Directions.HasFlag(Directions.South) && Directions.HasFlag(Directions.East))
            {
                Tiles.Add(new Tile(_game, "incorner",
                                   new Vector3(Coordinate.X - EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   0, this));
            }

            if (Directions.HasFlag(Directions.South) && Directions.HasFlag(Directions.West))
            {
                Tiles.Add(new Tile(_game, "incorner",
                                   new Vector3(Coordinate.X + EnvironmentService.TileSize, -1,
                                               Coordinate.Z - EnvironmentService.TileSize),
                                   -90, this));
            }
        }

        public float? ScanAndDraw(Ray r)
        {
            if (!_isTilesReady)
            {
                BuildModels();
                _isTilesReady = true;
            }

            foreach (var tile in Tiles)
            {
                if(r.Intersects(tile.Hitbox).HasValue)
                {
                    tile.Draw();
                    return r.Intersects(tile.Hitbox).Value;
                }
            }

            return null;
        }

        public void Draw()
        {
            if (!_isTilesReady)
            {
                BuildModels();
                _isTilesReady = true;
            }

            foreach (var tile in Tiles)
            {
                tile.Draw();
            }
        }

        public Vector3 GetNeighbour(Directions c2)
        {
            if (c2 == Directions.North)
                return new Vector3(SectorCoordinate.X, SectorCoordinate.Y, SectorCoordinate.Z + 1);
            else if (c2 == Directions.South)
                    return new Vector3(SectorCoordinate.X, SectorCoordinate.Y, SectorCoordinate.Z - 1);
            else if (c2 == Directions.East)
                return new Vector3(SectorCoordinate.X - 1, SectorCoordinate.Y, SectorCoordinate.Z);
            else if (c2 == Directions.West)
                return new Vector3(SectorCoordinate.X + 1, SectorCoordinate.Y, SectorCoordinate.Z);

            return Vector3.Zero;
        }
    }
}
